/* OS- and machine-dependent stuff for 4.[23] BSD UNIX */

/*
	FILE: UNIX.io.c
	
	Routines:
		asy_init()
		asy_stop()
		asy_speed()
		asy_output()
		get_asy()
		doasystat()
		asy_ioctl()
		asy_send()
		asy_tx()
		dir()
	Written or converted by Mikel Matthews, N9DVG, and others
	
*/

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/dir.h>
#include <sys/time.h>
#include "global.h"
#include "mbuf.h"
#include "internet.h"
#include "iface.h"
#include "unix.h"
#include "cmdparse.h"

struct asy Asy[ASY_MAX];

struct iface *Ifaces;

struct sgttyb mysavetty;

/* Initialize asynch port "dev" */
int
asy_init(dev,iface,arg1,arg2,bufsize,mode)
int16 dev;
struct iface *iface;
char *arg1,*arg2;
unsigned bufsize;
int mode;
{
	register struct asy *ap;
	extern struct iface *Ifaces;
	struct sgttyb sgttyb;
	struct fifo *fp;
	
	if (dev >= Nasy)
		return(-1);

	ap = &Asy[dev];
	ap->tty = malloc(strlen(arg2)+1);
#ifdef notyet
	if(!strncmp(arg2,"/dev/ptyp",8)) {
	    slipstart(arg2);
	    sleep(2);
	    arg2[5] = 't';
	}
#endif

	strcpy(ap->tty, arg2);
	ap->iface = iface;
	/* Set up receiver FIFO */
	fp = &ap->fifo;
	if((fp->buf = malloc(bufsize)) == NULLCHAR){
		tprintf("asy%d: No space for rx buffer\n",dev);
		return;
	}
	fp->bufsize = bufsize;
	fp->wp = fp->rp = fp->buf;
	fp->cnt = 0;

	tprintf("asy_init: tty name = %s\n", ap->tty);
	if ((ap->IORser = open (ap->tty, O_RDWR|O_NDELAY, 0)) < 0)
	{
		perror ("Could not open device for reading");
		return (-1);
	}
	if ((ap->IOWser = open (ap->tty, O_WRONLY, 0)) < 0)
	{
		perror ("Could not open device for writing");
		return (-1);
	}

 /* 
  * get the stty structure and save it 
  */

	if (ioctl (ap->IORser, TIOCGETP, &mysavetty) < 0)
	{
		perror ("ioctl failed on device");
		return (-1);
	}

 /* 
  * copy over the structure 
  */

	sgttyb = mysavetty;
	sgttyb.sg_flags = RAW;
	sgttyb.sg_ispeed = sgttyb.sg_ospeed = B9600;

 /* 
  * set up the parity and enter RAW mode 
  */
	sgttyb.sg_flags = (RAW | ANYP | CRMOD);
 /* 
  * now set the modes and flags 
  */
	if (ioctl (ap->IORser, TIOCSETP, &sgttyb) < 0)
	{
		perror ("ioctl could not set parameters for IORser");
		return (-1);
	}
	if (ioctl (ap->IOWser, TIOCSETP, &sgttyb) < 0)
	{
		perror ("ioctl could not set parameters for IOWser");
		return (-1);
	}

	/* Enable SIGIO interrupt */
	fcntl(ap->IORser,F_SETFL, fcntl(ap->IORser,F_GETFL,0) | FASYNC);
	fcntl(ap->IORser,F_SETOWN,getpid());
	return (0);
}

slipstart(buf)
char *buf;
{
    FILE *fp;
    int i;
    if(fork())
	return;
    for(i=0; i < Nasy; ++i) {
	close(Asy[i].IORser);
	close(Asy[i].IOWser);
    }
    close(0);
    close(1);
    close(2);
    fp = fopen(buf,"w+");
    dup2(fileno(fp),0);
    dup2(fileno(fp),1);
    dup2(fileno(fp),2);
    execl("/usr/local/etc/sliplogin","sliplogin",NULLCHAR);
}

int
asy_stop(dev)
int dev;
{
}

/* Set asynch line speed */
int
asy_speed(dev,speed)
int16 dev;
int speed;
{
	struct sgttyb sgttyb;

	if(speed == 0 || dev >= Nasy)
		return(-1);

	if (ioctl (Asy[dev].IORser, TIOCGETP, &sgttyb) < 0)
	{
		perror ("ioctl could not get parameters");
		return (-1);
	}
	switch(speed)
	{
		case 0:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B0;
			break;
		case 50:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B50;
			break;
		case 75:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B75;
			break;
		case 110:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B110;
			break;
		case 134:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B134;
			break;
		case 150:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B150;
			break;
		case 200:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B200;
			break;
		case 300:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B300;
			break;
		case 600:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B600;
			break;
		case 1200:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B1200;
			break;
		case 1800:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B1800;
			break;
		case 2400:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B2400;
			break;
		case 4800:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B4800;
			break;
		case 9600:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = B9600;
			break;
		case 19200:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = EXTA;
			break;
		case 38400:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = EXTB;
			break;
		/*
		case EXTA:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = EXTA;
			break;
		case EXTB:
			sgttyb.sg_ispeed = sgttyb.sg_ospeed = EXTB;
			break;
		*/
		default:
			tprintf("asy_speed: Unknown speed (%d)\n", speed);
			break;
	}

	if (ioctl (Asy[dev].IORser, TIOCSETP, &sgttyb) < 0)
	{
		perror ("ioctl could not set parameters for IORser");
		return (-1);
	}
	if (ioctl (Asy[dev].IOWser, TIOCSETP, &sgttyb) < 0)
	{
		perror ("ioctl could not set parameters for IOWser");
		return (-1);
	}

	Asy[dev].speed = speed;
	return(0);
}
/* Send a buffer to serial transmitter */
asy_output(dev,buf,cnt)
unsigned dev;
char *buf;
unsigned short cnt;
{
	int i_state;

	if(dev >= Nasy)
		return(-1);
	i_state = dirps();
	if ( write(Asy[dev].IOWser, buf, (int)cnt) < cnt)
	{
		perror("asy_output");
		printf("asy_output: error in writing to device %d\n", dev);
		restore (i_state);
		return(-1);
	}
	restore (i_state);
	Asy[dev].txchar += cnt;
	return(0);
}

/* Blocking read from asynch line
 * Returns the character just read
 */
char
get_asy(dev)
int16 dev;
{
	int i_state;
	register struct fifo *fp;
	char c;

	fp = &Asy[dev].fifo;

	i_state = dirps();
	while(fp->cnt == 0)
	    pwait(fp);
	fp->cnt--;

	c = *fp->rp++;
	if(fp->rp >= &fp->buf[fp->bufsize])
		fp->rp = fp->buf;

	restore (i_state);
	return c;
}

asy_ioctl(Iface, argc, argv)
struct iface *Iface;
int	argc;
char	*argv[];
{
	if (argc < 1) {
		tprintf("%d\n", Asy[Iface->dev].speed);
		return 0;
	}
	return asy_speed(Iface->dev, atoi(argv[0]));
}

int
doasystat(argc,argv)
int argc;
char *argv[];
{
	register struct asy *asyp;

	for(asyp = Asy;asyp < &Asy[Nasy];asyp++)
	    if (asyp->tty != NULLCHAR)
		tprintf("%s (%s %d baud): rxchar %lu txchar %lu\n",
		 asyp->iface->name,asyp->tty,asyp->speed,asyp->rxchar,asyp->txchar);
	return 0;
}

/* Send a message on the specified serial line */
int
asy_send(dev,bp)
int dev;
struct mbuf *bp;
{
	if(dev < 0 || dev >= Nasy)
		return -1;
	/* The data is sent directly rather than queued for asy_tx() */
	while(bp != NULLBUF) {
		asy_output(dev,bp->data,bp->cnt);
   		/* Now do next buffer on chain */
		bp = free_mbuf(bp);
	}
	return 0;
}

/* Serial transmit process, common to all protocols */
void
asy_tx(dev,p1,p2)
int dev;
void *p1;
void *p2;
{
/* This function is not used. Instead, each process handles the output
 * by itself. This is the way it used to be, and it works satisfactory
 * on the SUN.
 */
}

/* Generate a directory listing by opening a pipe to /bin/ls.
 * If full == 1, give a full listing; else return just a list of names.
 */
FILE *
dir(path,full)
char *path;
int full;
{
	FILE *fp;
	char cmd[1024];

	if (path == NULLCHAR || path[0] == '\0')
		path = ".";

	if (full)
		sprintf(cmd,"/bin/ls -lLF %s", path);
	else
		sprintf(cmd, "/bin/ls -CLF %s", path);

	if ((fp = popen(cmd,"r")) == NULLFILE) {
		perror("popen");
		return NULLFILE;
	}

	return fp;
}
