/*			NOS porting kit for UNIX with LWP
 *				    900410
 *
 * Files are: bsdunix.c, unix.h, bsd_io.c, bsd_nit.c, bsd_tun.c, kernel.c,
 * proc.h, global.h, Makefile
 *
 * This runs on a SUN-4, SUN-3 with the LightWeight Process Library.
 * The NIT and TUN drivers are activated by defining PACKET in config.h but
 * you need to run NOS as root to use them. TUN is the more efficient, so
 * I recommend its usage over NIT.
 *
 * You need release 900312 or later of KA9Q NOS to use these files.
 *
 * You will probably have to increase the definition of MSPTICK in
 * timer.h. I suggest doubling its value.
 * Delete any lines that cause troubles in hardware.h.
 *
 * You need to add the following line to the assignment of Daemons[]
 * in the file config.c:
       "interval timer", 1024, ctick,
 * also add a
 	void ctick();
 * to the top of config.c. There will be other compiler errors too, since
 * NOS was written for Turbo-C. But you are expected to fix those yourself.
 * Delete all references to missing include-files, etc.
 * 
 * Written by Anders Klemets, SM0RGV, klemets@sics.se
 * (a few functions at the end were written by Mikel Matthews, N9DVG
 *  Jere Sandidge, K4FUM, and some others.)
 * I would appreciate to see any improvements you make.
 *
 */
#include <stdio.h>
#include <sys/types.h>
#include <lwp/lwp.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/param.h>
#ifdef NeXT
#include <sys/dir.h>
#define dirent direct
#else
#include <dirent.h>
#endif
int free(); 		/* This must be before we include global.h */
int (*xxfree)() = free; /* if we are not using alloc.c */
#include "global.h"
#include "config.h"
#include "iface.h"
#include "proc.h"
#include "timer.h"
#include "session.h"
#include "unix.h"
#include "hardware.h"

void inpint(),iostop();
int Tick, console;
struct sgttyb mysavetty, savecon;
int Intstk[1024]; /* The signal stack */
struct proc *ctickp = NULLPROC;
static struct proc *inpintp = NULLPROC;
static mon_t mid;

/* Interface list header */
struct iface *Ifaces;

/* Keyboard input buffer */
struct fifo Keyboard;

extern struct proc *Display;

int shell=0,intoff=0;

/* Called at startup time to set up console I/O, memory heap */
ioinit(mem)
int mem;
{
	struct sgttyb ttybuf;
    	struct sigstack ss;
	struct sigvec sv;
	
	/* We want unbuffered output */
	if(isatty(1)) {
	    freopen(ttyname(1),"w",stdout);
	    setbuf(stdout,NULLCHAR);

	    if ((console = open("/dev/tty",O_RDONLY|O_NDELAY)) < 0) {
		perror("open /dev/tty");
		exit(1);
	    }
	    ioctl(console,TIOCGETP,&ttybuf);
	    savecon = ttybuf;
	    ttybuf.sg_flags &= ~ECHO;
	    ttybuf.sg_flags |= CBREAK;
	    ioctl(console,TIOCSETP,&ttybuf);
	}
	else console = 0;

	(void) signal(SIGHUP, iostop);
	(void) signal(SIGINT, iostop);
	(void) signal(SIGQUIT, iostop);
	(void) signal(SIGTERM, iostop);

	grabcore(mem); /* A null function when alloc.c is not used */

	/* Initialize keyboard queue */
	Keyboard.bufsize = 256;
	Keyboard.buf = malloc (256);
	Keyboard.rp = Keyboard.wp = Keyboard.buf;
	/* Cache a couple of 8k stacks. This number is referred to
	   in kernel.c as well */
	lwp_setstkcache(8192,5);

	mon_create(&mid);
}

void
ctick(p, v1, v2)
int p;
void *v1, *v2;
{
     char i_state;
     struct timeval timout;
     newproc("interrupt handler",1024,inpint,0,NULL,NULL);
     /* We will awaken the timer process at every MSPTICK milliseconds.
	You may try to tweak its definiton in timer.h */
     timout.tv_sec = 0;
     timout.tv_usec = MSPTICK * 1000;
     ctickp = Curproc;
/*     lwp_setpri(SELF,MINPRIO+1); */
     while(1) {
	  lwp_sleep(&timout);
	  Curproc = ctickp;
	  i_state = dirps();
	  ++Tick;
	  psignal(&Tick,1);
	  restore(i_state);
     }
}

/* Called just before exiting to restore console state */
void
iostop()
{
	while(Ifaces != NULLIF){
		if(Ifaces->stop != NULLFP)
			(*Ifaces->stop)(Ifaces->dev);
		Ifaces = Ifaces->next;
	}
	ioctl(console,TIOCSETP,&savecon);
	pod_exit(0);
}


/* Disable all "interrupts", don't mask any signals however */
int dirps()
{
    int mask;
/*    if(ctickp != NULLPROC)
	 lwp_setpri(ctickp->tid,MINPRIO);
    if(inpintp != NULLPROC)
	 lwp_setpri(inpintp->tid,MINPRIO); */
    if (intoff)
	 return 0;
    mon_enter(mid);
    intoff = 1;
    return 1;
}

/* Restore the "interrupts" */
void restore (mask)
    char mask;
{
     if (mask) {
	intoff = 0;
/*	if(ctickp != NULLPROC)
	     lwp_setpri(ctickp->tid,MINPRIO+1);
	if(inpintp != NULLPROC)
	     lwp_setpri(inpintp->tid,MINPRIO+1); */
	mon_exit(mid);
   }
}

/* Return the interrupt state */
int istate()
{
     return !intoff;
}

/* Input interrupt handler.
   This function is called whenever there is input from stdin, the
   asynchronus interfaces or the NIT driver. But unfortunately only if their
   respective input queues were empty.
   Speeding up this function would probably enhance the general performance
   of the program.
*/
void
inpint(p, v1, v2)
int p;
void *v1, *v2;
{
	int i,ok,r,cnt;
	struct timeval timeout;
	fd_set readfds;
	thread_t agt;
	eventinfo_t argbuf;
	char evbuf[100], *dummy, i_state;
	
	inpintp = Curproc;
/*	lwp_setpri(SELF,MINPRIO+1); */

	/* Create an agent for the SIGIO signal */
	agt_create(&agt, SIGIO, evbuf);
	     
	/* This will send a SIGIO signal whenever a key is pressed and
	   the stdin queue is empty. It would be better if the signal would
	   always be sent, regardless of the status of the queue. */
	if(isatty(console)) {
	    fcntl(console,F_SETFL, fcntl(console,F_GETFL,0) | FASYNC);
	    fcntl(console,F_SETOWN,getpid());
	}
	/* We use select() to se which file descriptors are ready for reading.
	   The syntax of select() differs between 4.2 and 4.3 BSD.
	 */
	timeout.tv_sec = 0;
	timeout.tv_usec = 35;
	
	/* This loop exits first when there is nothing to ready from the
	   asynchronus interfaces. */
	while (1) {
	     msg_recv(&agt,(caddr_t)&argbuf,&i,(caddr_t)dummy,&r,
		      INFINITY);
	     Curproc = inpintp;
	     FD_ZERO (&readfds);
	    for (i=0; i<Nasy; ++i)
		FD_SET(Asy[i].IORser,&readfds);
#ifdef PACKET
/* This is only for the NIT and Tun drivers */
	    for (i=0; i<Nnit; ++i)
		FD_SET(Nitdrvr[i].IOser,&readfds);
	    for (i=0; i<Ntun; ++i)
		FD_SET(Tundrvr[i].IOser,&readfds);
#endif
	    if(isatty(console))
		FD_SET(console,&readfds);
	     msg_reply(agt); /* Reply before we read anything */
	    ok = select(NOFILE,&readfds,(fd_set *)0,(fd_set *)0,&timeout);
	     i_state = dirps();
	     cnt = 0;
	    if (ok > 0) {
#ifdef PACKET
		for (i=0; i<Nnit; ++i)
		    if (FD_ISSET(Nitdrvr[i].IOser,&readfds))
			psignal(&Nitdrvr[i],1);
		for (i=0; i<Ntun; ++i)
		    if (FD_ISSET(Tundrvr[i].IOser,&readfds))
			psignal(&Tundrvr[i],1);
#endif
		/* Now scan the asynchronus interfaces. Remember, this is
		   supposed to be efficient. */
		for (i=0; i<Nasy; ++i)
		    if (FD_ISSET(Asy[i].IORser,&readfds)) {
			if (Asy[i].fifo.cnt == Asy[i].fifo.bufsize)
			    continue; /* No room */
			/* Try to read as much as we are allowed to */
			r = read(Asy[i].IORser, Asy[i].fifo.wp,
				 Asy[i].fifo.rp > Asy[i].fifo.wp ?
				 Asy[i].fifo.rp - Asy[i].fifo.wp :
				 &Asy[i].fifo.buf[Asy[i].fifo.bufsize] - Asy[i].fifo.wp);
			if (r <= 0)
			    continue;
			cnt += r;
			Asy[i].rxchar += r;
			Asy[i].fifo.wp += r;
			Asy[i].fifo.cnt += r;
			if(Asy[i].fifo.wp >= &Asy[i].fifo.buf[Asy[i].fifo.bufsize])
			    /* Wrap around */
			    Asy[i].fifo.wp = Asy[i].fifo.buf;
			/* Signal "asy_rx" that we have got new data. */
			psignal(&Asy[i].fifo,1);
		    }
	    }
	    /* This function would be a monster if I didn't split it up
	       into a special part for keyboard input. */
	    if (ok > 0 && FD_ISSET(console,&readfds))
		kbraw();
   	/* Now do it all over again since we might have received more input */
	restore(i_state);
	}
}
void dummyin(p,v1,v2)
int p;
void *v1,*v2;
{
     int ok;
     struct timeval timeout;
     fd_set readfds;
     timeout.tv_sec = 0;
	
	inpintp = Curproc;
	while (1) {
	     timeout.tv_usec = 100000;
	     lwp_sleep(&timeout);
	     Curproc = inpintp;
	     FD_ZERO (&readfds);
	    if(isatty(console))
		FD_SET(console,&readfds);
	     timeout.tv_usec = 35;
	    ok = select(NOFILE,&readfds,(fd_set *)0,(fd_set *)0,&timeout);
	     if (ok > 0)
		  kbraw();
	}
}

/* Read characters from stdin. Called from the generic interrupt handler. */
int
kbraw()
{
    int cnt;

    if (shell)
	 /* We are executing a shell right now */
	return 0;

    if (Keyboard.cnt == Keyboard.bufsize)
	return 0;
    /* Try to read as much as we are allowed to */
    cnt = read(console, Keyboard.wp,
	     Keyboard.rp > Keyboard.wp ? Keyboard.rp - Keyboard.wp :
	     Keyboard.buf + Keyboard.bufsize - Keyboard.wp);
    if (cnt <= 0)
	return 0;
    Keyboard.cnt += cnt;
    Keyboard.wp += cnt;
    if(Keyboard.wp >= Keyboard.buf + Keyboard.bufsize)
	Keyboard.wp = Keyboard.buf;
    psignal(&Keyboard,1);
    return cnt;
}
/* Blocking read from the keyboard */
int
kbread()
{
	int i_state;
	char c;

	i_state = dirps();
	while(Keyboard.cnt == 0)
		pwait (&Keyboard);
	c = *Keyboard.rp++;
	if(Keyboard.rp == &Keyboard.buf[Keyboard.bufsize])
		Keyboard.rp = Keyboard.buf;
	Keyboard.cnt--;
	restore(i_state);
	
	if (c == 0x7f) /* DEL key */
	    c = '\b';
	if (c == '\n') /* Emulate MS/DOS. SysV would reset the ICRNL flag */
	    c = '\r';
	return c;
}
systick()
{
}

sysreset()
{
}

/* This function is only needed if you DO use alloc.c */
unsigned long coreleft()
{
    struct rlimit rlp;
    extern end;
    getrlimit(RLIMIT_CORE,&rlp);
    return ((unsigned long) (rlp.rlim_cur - end));
}
/* Sleep gracefully until the next interrupt. */
giveup()
{
    restore(1); /* Interrupts must be enabled */
    sigpause(0);
    dirps();
}
/* To be removed if alloc.c is to be used */
memstat()
{
    return 0;
}
grabcore(mem)
{
}

/* Note: make sure that there is no previous definition of tolower() */
char tolower (c)
    char c;
{
    return ((c >= 'A' && c <= 'Z') ? c + 'a' - 'A' : c);
}

FILE *
tmpfile()
{
	FILE *tmp;
	char *mktemp();
	char *ptr = "SMTPXXXXXX";
	char *name;
	name = mktemp(ptr);
	if ( ( tmp = fopen(name, "w+") ) == NULL)
	{
		tprintf("tmpfile: counld not create temp file.\n");
		return(NULL);
	}
	(void) unlink(name);
	return ( tmp );
}

int stxrdy(dev)
{
    return 1;
}

/* wildcard filename lookup */
filedir(name, times, ret_str)
char	*name;
int	times;
char	*ret_str;
{
	static char	dname[128], fname[128];
	static DIR *dirp = NULL;
	struct dirent *dp;
	struct stat sbuf;
	char	*cp, temp[128];

	/*
	 * Make sure that the NULL is there in case we don't find anything
	 */
	ret_str[0] = '\0';

	if (times == 0) {
		/* default a null name to *.* */
		if (name == NULL || *name == '\0')
			name = "*.*";
		/* split path into directory and filename */
		if ((cp = strrchr(name, '/')) == NULL) {
			strcpy(dname, ".");
			strcpy(fname, name);
		} else {
			strcpy(dname, name);
			dname[cp - name] = '\0';
			strcpy(fname, cp + 1);
			/* root directory */
			if (dname[0] == '\0')
				strcpy(dname, "/");
			/* trailing '/' */
			if (fname[0] == '\0')
				strcpy(fname, "*.*");
		}
		/* close directory left over from another call */
		if (dirp != NULL)
			closedir(dirp);
		/* open directory */
		if ((dirp = opendir(dname)) == NULL) {
			tprintf("Could not open DIR (%s)\n", dname);
			return;
		}
	} else {
		/* for people who don't check return values */
		if (dirp == NULL)
			return;
	}

	/* scan directory */
	while ((dp = readdir(dirp)) != NULL) {
		/* test for name match */
/*		if (wildmat(dp->d_name, fname)) {*/
		if (!strcmp(dp->d_name, fname) || (fname[0] == '*' &&
		    !strcmp(&dp->d_name[strlen(dp->d_name) - strlen(fname)+1],
		    &fname[1]))) { /* ...when we do not use wildmat */
			/* test for regular file */
			sprintf(temp, "%s/%s", dname, dp->d_name);
			if (stat(temp, &sbuf) < 0)
				continue;
			if ((sbuf.st_mode & S_IFREG) == 0)
				continue;
			strcpy(ret_str, dp->d_name);
			break;
		}
	}

	/* close directory if we hit the end */
	if (dp == NULL) {
		closedir(dirp);
		dirp = NULL;
	}
}
doshell(argc, argv)
char **argv;
{
	register int pid, pid1, i;
	void (*savi)();
	int rc;
	char str[256], *cp, *getenv();
	struct sgttyb ttybuf;
	extern struct sgttyb savecon;
	shell = 1;

	ioctl(console, TIOCGETP, &ttybuf);
	ioctl(console, TIOCSETP, &savecon);

	str[0] = '\0';
	for (i = 1; i < argc; i++) {
		strcat(str, argv[i]);
		strcat(str, " ");
	}
	if ((cp = getenv("SHELL")) == NULLCHAR)
		cp = "/bin/sh";
	if ((pid = fork()) == 0) {
		seteuid(getuid()); /* Restore original user ID */
		if (argc > 1)
			(void)execl(cp, cp, "-c", str, 0);
		else
			(void)execl(cp, cp, (char *)0,(char *)0,0);
		tprintf("Can't execl() (%d)\n", errno);
		exit(1);
	} else if (pid == -1) {
		tprintf("Can't fork() (%d)\n", errno);
		rc = -1;
	} else {
		savi = signal(SIGINT, SIG_IGN);
		while ((pid1 = wait(&rc)) != pid && pid1 != -1)
			;	
		signal(SIGINT, savi);
	}
	ioctl(console, TIOCSETP, &ttybuf);
	shell = 0;
	return (rc);
}

dodir(argc, argv)
char **argv;
{
	register int i;
	char str[256];

	strcpy(str, "ls -lL ");
	for (i = 1; i < argc; i++) {
		strcat(str, argv[i]);
		strcat(str, " ");
	}

	return system(str);
}
    
/* Terminal printf that counts newlines. At end of the page,
 * prints "--More--" and waits for a keystroke before continuing.
 */
void
display(i,v1,v2)
int i;
void *v1;
void *v2;
{
	int c;
	struct session *sp;

	/* This is very tricky code. Because the value of "Current" can
	 * change any time we do a pwait, we have to be careful to detect
	 * any change and go back and start again.
	 */
	for(;;){
		sp = Current;

		if(sp->morewait){
			pwait(&sp->row);
			if(sp != Current || sp->row <= 0){
				/* Current changed value, or the user
				 * hasn't really hit a key
				 */
				continue;
			}
			/* Erase the prompt */
			printf("\r        \r");
		}
		sp->morewait = 0;
		if((c = recvchar(sp->output)) == -1){
			/* the alert() in swapscreen will cause this to
			 * return -1 when current changes
			 */
			pwait(NULL);	/* Prevent a nasty loop */
			continue;
		}
		putchar(c);
		if(sp->record != NULLFILE)
			fputc(c,sp->record);
		if(sp->flowmode && c == '\n' && --sp->row <= 0){
			printf("--More--");
			sp->morewait = 1;
		}
	}
}
void swapscreen(old, new)
struct session *old;
struct session *new;
{
	if(old == new)
		return;	/* Nothing to do */

	fflush(stdout);
	if(old != NULLSESSION){
		/* Save old screen */
	}
	if(new != NULLSESSION){
		/* Load new screen */
	}
	alert(Display,1);	/* Wake him up */
}
void
newscreen(sp)
struct session *sp;
{
	if(sp != NULLSESSION)
		sp->screen = (struct screen *)callocw(1,sizeof(struct screen));
}
void
freescreen(sp)
struct session *sp;
{
	if(sp == NULLSESSION || sp->screen == NULLSCREEN)
		return;
	if(sp->screen->save != NULLCHAR)
		free(sp->screen->save);
	free((char *)sp->screen);
}
#ifdef NeXT
/* Copy a string to a malloc'ed buffer */
char *
strdup(s)
const char *s;
{
	register char *out;
	register int len;

	if(s == NULLCHAR)
		return NULLCHAR;
	len = strlen(s);
	out = mallocw(len+1);
	/* This is probably a tad faster than strcpy, since we know the len */
	memcpy(out,s,len);
	out[len] = '\0';
	return out;
}

/* Case-insensitive string comparison */
strnicmp(a,b,n)
register char *a,*b;
register int n;
{
	char a1,b1;

	while(n-- != 0 && (a1 = *a++) != '\0' && (b1 = *b++) != '\0'){
		if(a1 == b1)
			continue;	/* No need to convert */
		if(isupper(a1))
			a1 = tolower(a1);
		if(isupper(b1))
			b1 = tolower(b1);
		if(a1 == b1)
			continue;	/* NOW they match! */
		if(a1 > b1)
			return 1;
		if(a1 < b1)
			return -1;
	}
	return 0;
}
/* Case-insensitive string comparison */
stricmp(a,b)
register char *a,*b;
{
	char a1,b1;

	while((a1 = *a++) != '\0' && (b1 = *b++) != '\0'){
		if(a1 == b1)
			continue;	/* No need to convert */
		if(isupper(a1))
			a1 = tolower(a1);
		if(isupper(b1))
			b1 = tolower(b1);
		if(a1 == b1)
			continue;	/* NOW they match! */
		if(a1 > b1)
			return 1;
		if(a1 < b1)
			return -1;
	}
	return 0;
}
#endif	/* NeXT */
